@@ -11,20 +11,18 @@
 // along with this software.
 // If not, see <https://opensource.org/licenses/Apache-2.0>.
 
-#![recursion_limit = "256"]
+//! Bitcoin SLIP-132 standard implementation for parsing custom xpub/xpriv key
+//! formats
+
 // Coding conventions
-#![deny(
-    non_upper_case_globals,
-    non_camel_case_types,
-    non_snake_case,
-    unused_mut,
-    unused_imports,
-    dead_code,
-    //missing_docs
-)]
+#![recursion_limit = "256"]
+#![deny(dead_code, missing_docs, warnings)]
 
 #[macro_use]
 extern crate amplify;
+#[cfg(feature = "strict_encoding")]
+#[macro_use]
+extern crate strict_encoding;
 
 #[cfg(feature = "serde")]
 #[macro_use]
@@ -36,9 +34,7 @@ use std::fmt::Debug;
 use std::str::FromStr;
 
 use bitcoin::util::base58;
-use bitcoin::util::bip32::{
-    self, ChildNumber, DerivationPath, ExtendedPrivKey, ExtendedPubKey,
-};
+use bitcoin::util::bip32::{self, ChildNumber, DerivationPath, ExtendedPrivKey, ExtendedPubKey};
 use bitcoin::Network;
 
 /// Magical version bytes for xpub: bitcoin mainnet public key for P2PKH or P2SH
@@ -101,61 +97,71 @@ pub const VERSION_MAGIC_VPUB_MULTISIG: [u8; 4] = [0x02, 0x57, 0x54, 0x83];
 pub const VERSION_MAGIC_VPRV_MULTISIG: [u8; 4] = [0x02, 0x57, 0x50, 0x48];
 
 /// Extended public and private key processing errors
-// TODO: Derive `PartialOrd`, `Ord`, and `Hash`, once `base58::Error` will have
-#[derive(Clone, PartialEq, Eq, Debug, Display, From, Error)]
+#[derive(
+    Clone, PartialEq, Eq, PartialOrd, Ord, Hash, Debug, Display, From, Error
+)]
 #[display(doc_comments)]
 pub enum Error {
-    /// Error in BASE58 key encoding
+    /// error in BASE58 key encoding. Details: {0}
     #[from]
     Base58(base58::Error),
 
-    /// A pk->pk derivation was attempted on a hardened key
+    /// error in hex key encoding. Details: {0}
+    #[from]
+    Hex(bitcoin::hashes::hex::Error),
+
+    /// pk->pk derivation was attempted on a hardened key.
     CannotDeriveFromHardenedKey,
 
-    /// A child number was provided ({0}) that was out of range
+    /// child number {0} is out of range.
     InvalidChildNumber(u32),
 
-    /// Invalid child number format.
+    /// invalid child number format.
     InvalidChildNumberFormat,
 
-    /// Invalid derivation path format.
+    /// invalid derivation path format.
     InvalidDerivationPathFormat,
 
-    /// Unknown version magic bytes
+    /// unknown version magic bytes {0:#06X?}
     UnknownVersion([u8; 4]),
 
-    /// Encoded extended key data has wrong length
+    /// encoded extended key data has wrong length {0}
     WrongExtendedKeyLength(usize),
 
-    /// Unrecognized or unsupported extended key prefix (please check SLIP 32
+    /// unrecognized or unsupported extended key prefix (please check SLIP 32
     /// for possible values)
     UnknownSlip32Prefix,
 
-    /// Failure in rust bitcoin library
+    /// failure in rust bitcoin library
     InternalFailure,
 }
 
+#[cfg(feature = "strict_encoding")]
+impl strict_encoding::StrictEncode for Error {
+    fn strict_encode<E: std::io::Write>(&self, _: E) -> Result<usize, strict_encoding::Error> {
+        unreachable!("StrictEncode for slip132::Error is a dummy required by miniscript")
+    }
+}
+
+#[cfg(feature = "strict_encoding")]
+impl strict_encoding::StrictDecode for Error {
+    fn strict_decode<D: std::io::Read>(_: D) -> Result<Self, strict_encoding::Error> {
+        unreachable!("StrictDecode for slip132::Error is a dummy required by miniscript")
+    }
+}
+
 impl From<bip32::Error> for Error {
     fn from(err: bip32::Error) -> Self {
         match err {
-            bip32::Error::CannotDeriveFromHardenedKey => {
-                Error::CannotDeriveFromHardenedKey
-            }
-            bip32::Error::InvalidChildNumber(no) => {
-                Error::InvalidChildNumber(no)
-            }
-            bip32::Error::InvalidChildNumberFormat => {
-                Error::InvalidChildNumberFormat
-            }
-            bip32::Error::InvalidDerivationPathFormat => {
-                Error::InvalidDerivationPathFormat
-            }
-            bip32::Error::Ecdsa(_) => Error::InternalFailure,
+            bip32::Error::CannotDeriveFromHardenedKey => Error::CannotDeriveFromHardenedKey,
+            bip32::Error::InvalidChildNumber(no) => Error::InvalidChildNumber(no),
+            bip32::Error::InvalidChildNumberFormat => Error::InvalidChildNumberFormat,
+            bip32::Error::InvalidDerivationPathFormat => Error::InvalidDerivationPathFormat,
+            bip32::Error::Secp256k1(_) => Error::InternalFailure,
             bip32::Error::UnknownVersion(ver) => Error::UnknownVersion(ver),
-            bip32::Error::WrongExtendedKeyLength(len) => {
-                Error::WrongExtendedKeyLength(len)
-            }
+            bip32::Error::WrongExtendedKeyLength(len) => Error::WrongExtendedKeyLength(len),
             bip32::Error::Base58(err) => Error::Base58(err),
+            bip32::Error::Hex(err) => Error::Hex(err),
         }
     }
 }
@@ -168,6 +174,23 @@ impl From<bip32::Error> for Error {
 #[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash, Debug)]
 pub struct KeyVersion([u8; 4]);
 
+#[cfg(feature = "strict_encoding")]
+impl strict_encoding::StrictEncode for KeyVersion {
+    fn strict_encode<E: std::io::Write>(&self, mut e: E) -> Result<usize, strict_encoding::Error> {
+        e.write_all(&self.0)?;
+        Ok(4)
+    }
+}
+
+#[cfg(feature = "strict_encoding")]
+impl strict_encoding::StrictDecode for KeyVersion {
+    fn strict_decode<D: std::io::Read>(mut d: D) -> Result<Self, strict_encoding::Error> {
+        let mut bytes = [0u8; 4];
+        d.read_exact(&mut bytes)?;
+        Ok(Self(bytes))
+    }
+}
+
 /// Trait which must be implemented by helpers which do construction,
 /// interpretation, verification and cross-conversion of extended public and
 /// private key version magic bytes from [`KeyVersion`]
@@ -191,99 +214,74 @@ pub trait VersionResolver:
 
     /// Detects whether provided version corresponds to an extended public key.
     /// Returns `None` if the version is not recognized/unknown to the resolver.
-    fn is_pub(_: &KeyVersion) -> Option<bool> {
-        return None;
-    }
+    fn is_pub(_: &KeyVersion) -> Option<bool> { None }
 
     /// Detects whether provided version corresponds to an extended private key.
     /// Returns `None` if the version is not recognized/unknown to the resolver.
-    fn is_prv(_: &KeyVersion) -> Option<bool> {
-        return None;
-    }
+    fn is_prv(_: &KeyVersion) -> Option<bool> { None }
 
     /// Detects network used by the provided key version bytes.
     /// Returns `None` if the version is not recognized/unknown to the resolver.
-    fn network(_: &KeyVersion) -> Option<Self::Network> {
-        return None;
-    }
+    fn network(_: &KeyVersion) -> Option<Self::Network> { None }
 
     /// Detects application scope defined by the provided key version bytes.
     /// Application scope is a types of scriptPubkey descriptors in which given
     /// extended public/private keys can be used.
     /// Returns `None` if the version is not recognized/unknown to the resolver.
-    fn application(_: &KeyVersion) -> Option<Self::Application> {
-        return None;
-    }
+    fn application(_: &KeyVersion) -> Option<Self::Application> { None }
 
     /// Returns BIP 32 derivation path for the provided key version.
     /// Returns `None` if the version is not recognized/unknown to the resolver.
-    fn derivation_path(_: &KeyVersion) -> Option<DerivationPath> {
-        return None;
-    }
+    fn derivation_path(_: &KeyVersion, _: Option<ChildNumber>) -> Option<DerivationPath> { None }
 
     /// Converts version into version corresponding to an extended public key.
     /// Returns `None` if the resolver does not know how to perform conversion.
-    fn make_pub(_: &KeyVersion) -> Option<KeyVersion> {
-        return None;
-    }
+    fn make_pub(_: &KeyVersion) -> Option<KeyVersion> { None }
 
     /// Converts version into version corresponding to an extended private key.
     /// Returns `None` if the resolver does not know how to perform conversion.
-    fn make_prv(_: &KeyVersion) -> Option<KeyVersion> {
-        return None;
-    }
+    fn make_prv(_: &KeyVersion) -> Option<KeyVersion> { None }
 }
 
 impl KeyVersion {
     /// Detects whether provided version corresponds to an extended public key.
     /// Returns `None` if the version is not recognized/unknown to the resolver.
-    pub fn is_pub<R: VersionResolver>(&self) -> Option<bool> {
-        R::is_pub(&self)
-    }
+    pub fn is_pub<R: VersionResolver>(&self) -> Option<bool> { R::is_pub(self) }
 
     /// Detects whether provided version corresponds to an extended private key.
     /// Returns `None` if the version is not recognized/unknown to the resolver.
-    pub fn is_prv<R: VersionResolver>(&self) -> Option<bool> {
-        R::is_prv(&self)
-    }
+    pub fn is_prv<R: VersionResolver>(&self) -> Option<bool> { R::is_prv(self) }
 
     /// Detects network used by the provided key version bytes.
     /// Returns `None` if the version is not recognized/unknown to the resolver.
-    pub fn network<R: VersionResolver>(&self) -> Option<R::Network> {
-        R::network(&self)
-    }
+    pub fn network<R: VersionResolver>(&self) -> Option<R::Network> { R::network(self) }
 
     /// Detects application scope defined by the provided key version bytes.
     /// Application scope is a types of scriptPubkey descriptors in which given
     /// extended public/private keys can be used.
     /// Returns `None` if the version is not recognized/unknown to the resolver.
-    pub fn application<R: VersionResolver>(&self) -> Option<R::Application> {
-        R::application(&self)
-    }
+    pub fn application<R: VersionResolver>(&self) -> Option<R::Application> { R::application(self) }
 
     /// Returns BIP 32 derivation path for the provided key version.
     /// Returns `None` if the version is not recognized/unknown to the resolver.
     pub fn derivation_path<R: VersionResolver>(
         &self,
+        account: Option<ChildNumber>,
     ) -> Option<DerivationPath> {
-        R::derivation_path(&self)
+        R::derivation_path(self, account)
     }
 
     /// Converts version into version corresponding to an extended public key.
     /// Returns `None` if the resolver does not know how to perform conversion.
-    pub fn try_to_pub<R: VersionResolver>(&self) -> Option<KeyVersion> {
-        R::make_pub(&self)
-    }
+    pub fn try_to_pub<R: VersionResolver>(&self) -> Option<KeyVersion> { R::make_pub(self) }
 
     /// Converts version into version corresponding to an extended private key.
     /// Returns `None` if the resolver does not know how to perform conversion.
-    pub fn try_to_prv<R: VersionResolver>(&self) -> Option<KeyVersion> {
-        R::make_prv(&self)
-    }
+    pub fn try_to_prv<R: VersionResolver>(&self) -> Option<KeyVersion> { R::make_prv(self) }
 }
 
 /// Default resolver knowing native [`bitcoin::network::constants::Network`]
-/// and BIP 32 and SLIP 132-defined key applications with [`KeyApplications`]
+/// and BIP 32 and SLIP 132-defined key applications with [`KeyApplication`]
 #[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash, Debug)]
 pub struct DefaultResolver;
 
@@ -294,42 +292,43 @@ pub struct DefaultResolver;
     derive(Serialize, Deserialize),
     serde(crate = "serde_crate")
 )]
+#[cfg_attr(feature = "strict_encoding", derive(StrictEncode, StrictDecode))]
 #[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Debug, Display)]
 #[non_exhaustive]
 pub enum KeyApplication {
     /// xprv/xpub: keys that can be used for P2PKH and multisig P2SH
     /// scriptPubkey descriptors.
-    #[display("hashed")]
-    #[cfg_attr(feature = "serde", serde(rename = "hashed"))]
+    #[display("BIP44")]
+    #[cfg_attr(feature = "serde", serde(rename = "bip44"))]
     Hashed,
 
     /// zprv/zpub: keys that can be used for P2WPKH scriptPubkey descriptors
-    #[display("segwit")]
-    #[cfg_attr(feature = "serde", serde(rename = "segwit"))]
+    #[display("BIP84")]
+    #[cfg_attr(feature = "serde", serde(rename = "bip84"))]
     SegWit,
 
     /// Zprv/Zpub: keys that can be used for multisig P2WSH scriptPubkey
     /// descriptors
-    #[display("segwit-multisig")]
-    #[cfg_attr(feature = "serde", serde(rename = "segwit-multisig"))]
+    #[display("BIP48-native")]
+    #[cfg_attr(feature = "serde", serde(rename = "bip48-native"))]
     SegWitMiltisig,
 
     /// yprv/ypub: keys that can be used for P2WPKH-in-P2SH scriptPubkey
     /// descriptors
-    #[display("nested")]
-    #[cfg_attr(feature = "serde", serde(rename = "nested"))]
+    #[display("BIP49")]
+    #[cfg_attr(feature = "serde", serde(rename = "bip49"))]
     Nested,
 
     /// Yprv/Ypub: keys that can be used for multisig P2WSH-in-P2SH
     /// scriptPubkey descriptors
-    #[display("nested-multisig")]
-    #[cfg_attr(feature = "serde", serde(rename = "nested-multisig"))]
+    #[display("BIP48-nested")]
+    #[cfg_attr(feature = "serde", serde(rename = "bip48-nested"))]
     NestedMultisig,
 }
 
 /// Unknown string representation of [`KeyApplication`] enum
 #[derive(
-    Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash, Debug, Display, Error,
+    Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash, Debug, Display, Error
 )]
 #[display(doc_comments)]
 pub struct UnknownKeyApplicationError;
@@ -339,17 +338,76 @@ impl FromStr for KeyApplication {
 
     fn from_str(s: &str) -> Result<Self, Self::Err> {
         Ok(match s.to_lowercase().as_str() {
-            "pkh" => KeyApplication::Hashed,
-            "sh" => KeyApplication::Hashed,
-            "wpkh" => KeyApplication::SegWit,
-            "wsh" => KeyApplication::SegWitMiltisig,
-            "wpkh-sh" => KeyApplication::Nested,
-            "wsh-sh" => KeyApplication::NestedMultisig,
-            _ => Err(UnknownKeyApplicationError)?,
+            "bip44" => KeyApplication::Hashed,
+            "bip84" => KeyApplication::SegWit,
+            "bip48-native" => KeyApplication::SegWitMiltisig,
+            "bip49" => KeyApplication::Nested,
+            "bip48-nested" => KeyApplication::NestedMultisig,
+            _ => return Err(UnknownKeyApplicationError),
         })
     }
 }
 
+impl KeyApplication {
+    /// Enumerates all application variants    
+    pub const ALL: [KeyApplication; 5] = [
+        KeyApplication::Hashed,
+        KeyApplication::SegWit,
+        KeyApplication::SegWitMiltisig,
+        KeyApplication::Nested,
+        KeyApplication::NestedMultisig,
+    ];
+
+    /// Deduces application variant corresponding to the provided derivation
+    /// path, if possible.
+    pub fn from_derivation_path(path: DerivationPath) -> Option<KeyApplication> {
+        let path: Vec<_> = path.into();
+        for application in &[
+            KeyApplication::Hashed,
+            KeyApplication::SegWit,
+            KeyApplication::SegWitMiltisig,
+            KeyApplication::Nested,
+            KeyApplication::NestedMultisig,
+        ] {
+            if let Some(standard) = application.to_derivation_path() {
+                let standard: Vec<_> = standard.into();
+                if standard.strip_prefix(path.as_slice()).is_some() {
+                    return Some(*application);
+                }
+            }
+        }
+        let bip48_purpose = ChildNumber::Hardened { index: 48 };
+        if path.len() >= 4 && path[0] == bip48_purpose {
+            match path[3] {
+                ChildNumber::Hardened { index: 1 } => Some(KeyApplication::NestedMultisig),
+                ChildNumber::Hardened { index: 2 } => Some(KeyApplication::SegWitMiltisig),
+                _ => None,
+            }
+        } else {
+            None
+        }
+    }
+
+    /// Constructs derivation path matching the provided application
+    pub fn to_derivation_path(&self) -> Option<DerivationPath> {
+        match self {
+            Self::Hashed => Some(DerivationPath::from(vec![
+                ChildNumber::Hardened { index: 44 },
+                ChildNumber::Hardened { index: 0 },
+            ])),
+            Self::Nested => Some(DerivationPath::from(vec![
+                ChildNumber::Hardened { index: 49 },
+                ChildNumber::Hardened { index: 0 },
+            ])),
+            Self::SegWit => Some(DerivationPath::from(vec![
+                ChildNumber::Hardened { index: 84 },
+                ChildNumber::Hardened { index: 0 },
+            ])),
+            _ => None,
+        }
+    }
+}
+
 impl KeyVersion {
     /// Tries to construct [`KeyVersion`] object from a byte slice. If byte
     /// slice length is not equal to 4, returns `None`
@@ -362,41 +420,37 @@ impl KeyVersion {
         Some(KeyVersion::from_bytes(bytes))
     }
 
-    /// Constructs [`KeyVersion`] from a fixed 4 bytes values
-    pub fn from_bytes(version_bytes: [u8; 4]) -> KeyVersion {
-        KeyVersion(version_bytes)
+    /// Constructs [`KeyVersion`] from a Base58-encoded extended key string.
+    ///
+    /// # Panics
+    /// If the string does not contain at least 5 characters.
+    #[inline]
+    pub fn from_xkey_str(key: &str) -> Result<KeyVersion, Error> {
+        let xkey = base58::from(key)?;
+        KeyVersion::from_slice(&xkey[..4]).ok_or(Error::UnknownSlip32Prefix)
     }
 
+    /// Constructs [`KeyVersion`] from a fixed 4 bytes values
+    pub fn from_bytes(version_bytes: [u8; 4]) -> KeyVersion { KeyVersion(version_bytes) }
+
     /// Constructs [`KeyVersion`] from a `u32`-representation of the version
     /// bytes (the representation must be in bing endian format)
-    pub fn from_u32(version: u32) -> KeyVersion {
-        KeyVersion(version.to_be_bytes())
-    }
+    pub fn from_u32(version: u32) -> KeyVersion { KeyVersion(version.to_be_bytes()) }
 
     /// Converts version bytes into `u32` representation in big endian format
-    pub fn to_u32(&self) -> u32 {
-        u32::from_be_bytes(self.0)
-    }
+    pub fn to_u32(&self) -> u32 { u32::from_be_bytes(self.0) }
 
     /// Returns slice representing internal version bytes
-    pub fn as_slice(&self) -> &[u8] {
-        &self.0
-    }
+    pub fn as_slice(&self) -> &[u8] { &self.0 }
 
     /// Returns internal representation of version bytes
-    pub fn as_bytes(&self) -> &[u8; 4] {
-        &self.0
-    }
+    pub fn as_bytes(&self) -> &[u8; 4] { &self.0 }
 
     /// Constructs 4-byte array containing version byte values
-    pub fn to_bytes(&self) -> [u8; 4] {
-        self.0
-    }
+    pub fn to_bytes(&self) -> [u8; 4] { self.0 }
 
     /// Converts into 4-byte array containing version byte values
-    pub fn into_bytes(self) -> [u8; 4] {
-        self.0
-    }
+    pub fn into_bytes(self) -> [u8; 4] { self.0 }
 }
 
 impl VersionResolver for DefaultResolver {
@@ -409,24 +463,12 @@ impl VersionResolver for DefaultResolver {
         is_priv: bool,
     ) -> KeyVersion {
         match (network, applicable_for, is_priv) {
-            (Network::Bitcoin, KeyApplication::Hashed, false) => {
-                KeyVersion(VERSION_MAGIC_XPUB)
-            }
-            (Network::Bitcoin, KeyApplication::Hashed, true) => {
-                KeyVersion(VERSION_MAGIC_XPRV)
-            }
-            (Network::Bitcoin, KeyApplication::Nested, false) => {
-                KeyVersion(VERSION_MAGIC_YPUB)
-            }
-            (Network::Bitcoin, KeyApplication::Nested, true) => {
-                KeyVersion(VERSION_MAGIC_YPRV)
-            }
-            (Network::Bitcoin, KeyApplication::SegWit, false) => {
-                KeyVersion(VERSION_MAGIC_ZPUB)
-            }
-            (Network::Bitcoin, KeyApplication::SegWit, true) => {
-                KeyVersion(VERSION_MAGIC_ZPRV)
-            }
+            (Network::Bitcoin, KeyApplication::Hashed, false) => KeyVersion(VERSION_MAGIC_XPUB),
+            (Network::Bitcoin, KeyApplication::Hashed, true) => KeyVersion(VERSION_MAGIC_XPRV),
+            (Network::Bitcoin, KeyApplication::Nested, false) => KeyVersion(VERSION_MAGIC_YPUB),
+            (Network::Bitcoin, KeyApplication::Nested, true) => KeyVersion(VERSION_MAGIC_YPRV),
+            (Network::Bitcoin, KeyApplication::SegWit, false) => KeyVersion(VERSION_MAGIC_ZPUB),
+            (Network::Bitcoin, KeyApplication::SegWit, true) => KeyVersion(VERSION_MAGIC_ZPRV),
             (Network::Bitcoin, KeyApplication::NestedMultisig, false) => {
                 KeyVersion(VERSION_MAGIC_YPUB_MULTISIG)
             }
@@ -439,30 +481,16 @@ impl VersionResolver for DefaultResolver {
             (Network::Bitcoin, KeyApplication::SegWitMiltisig, true) => {
                 KeyVersion(VERSION_MAGIC_ZPRV_MULTISIG)
             }
-            (_, KeyApplication::Hashed, false) => {
-                KeyVersion(VERSION_MAGIC_TPUB)
-            }
+            (_, KeyApplication::Hashed, false) => KeyVersion(VERSION_MAGIC_TPUB),
             (_, KeyApplication::Hashed, true) => KeyVersion(VERSION_MAGIC_TPRV),
-            (_, KeyApplication::Nested, false) => {
-                KeyVersion(VERSION_MAGIC_UPUB)
-            }
+            (_, KeyApplication::Nested, false) => KeyVersion(VERSION_MAGIC_UPUB),
             (_, KeyApplication::Nested, true) => KeyVersion(VERSION_MAGIC_UPRV),
-            (_, KeyApplication::SegWit, false) => {
-                KeyVersion(VERSION_MAGIC_VPUB)
-            }
+            (_, KeyApplication::SegWit, false) => KeyVersion(VERSION_MAGIC_VPUB),
             (_, KeyApplication::SegWit, true) => KeyVersion(VERSION_MAGIC_VPRV),
-            (_, KeyApplication::NestedMultisig, false) => {
-                KeyVersion(VERSION_MAGIC_UPUB_MULTISIG)
-            }
-            (_, KeyApplication::NestedMultisig, true) => {
-                KeyVersion(VERSION_MAGIC_UPRV_MULTISIG)
-            }
-            (_, KeyApplication::SegWitMiltisig, false) => {
-                KeyVersion(VERSION_MAGIC_VPUB_MULTISIG)
-            }
-            (_, KeyApplication::SegWitMiltisig, true) => {
-                KeyVersion(VERSION_MAGIC_VPRV_MULTISIG)
-            }
+            (_, KeyApplication::NestedMultisig, false) => KeyVersion(VERSION_MAGIC_UPUB_MULTISIG),
+            (_, KeyApplication::NestedMultisig, true) => KeyVersion(VERSION_MAGIC_UPRV_MULTISIG),
+            (_, KeyApplication::SegWitMiltisig, false) => KeyVersion(VERSION_MAGIC_VPUB_MULTISIG),
+            (_, KeyApplication::SegWitMiltisig, true) => KeyVersion(VERSION_MAGIC_VPRV_MULTISIG),
         }
     }
 
@@ -492,9 +520,7 @@ impl VersionResolver for DefaultResolver {
         }
     }
 
-    fn is_prv(kv: &KeyVersion) -> Option<bool> {
-        DefaultResolver::is_pub(kv).map(|v| !v)
-    }
+    fn is_prv(kv: &KeyVersion) -> Option<bool> { DefaultResolver::is_pub(kv).map(|v| !v) }
 
     fn network(kv: &KeyVersion) -> Option<Self::Network> {
         match kv.as_bytes() {
@@ -531,83 +557,94 @@ impl VersionResolver for DefaultResolver {
             &VERSION_MAGIC_YPUB_MULTISIG
             | &VERSION_MAGIC_YPRV_MULTISIG
             | &VERSION_MAGIC_UPUB_MULTISIG
-            | &VERSION_MAGIC_UPRV_MULTISIG => {
-                Some(KeyApplication::NestedMultisig)
-            }
+            | &VERSION_MAGIC_UPRV_MULTISIG => Some(KeyApplication::NestedMultisig),
             &VERSION_MAGIC_ZPUB | &VERSION_MAGIC_ZPRV | &VERSION_MAGIC_VPUB
             | &VERSION_MAGIC_VPRV => Some(KeyApplication::SegWit),
             &VERSION_MAGIC_ZPUB_MULTISIG
             | &VERSION_MAGIC_ZPRV_MULTISIG
             | &VERSION_MAGIC_VPUB_MULTISIG
-            | &VERSION_MAGIC_VPRV_MULTISIG => {
-                Some(KeyApplication::SegWitMiltisig)
-            }
+            | &VERSION_MAGIC_VPRV_MULTISIG => Some(KeyApplication::SegWitMiltisig),
             _ => None,
         }
     }
 
-    fn derivation_path(kv: &KeyVersion) -> Option<DerivationPath> {
+    fn derivation_path(kv: &KeyVersion, account: Option<ChildNumber>) -> Option<DerivationPath> {
         match kv.as_bytes() {
-            &VERSION_MAGIC_XPUB | &VERSION_MAGIC_XPRV => {
-                Some(DerivationPath::from(vec![
-                    ChildNumber::Hardened { index: 44 },
+            &VERSION_MAGIC_XPUB | &VERSION_MAGIC_XPRV => Some(vec![
+                ChildNumber::Hardened { index: 44 },
+                ChildNumber::Hardened { index: 0 },
+            ]),
+            &VERSION_MAGIC_TPUB | &VERSION_MAGIC_TPRV => Some(vec![
+                ChildNumber::Hardened { index: 44 },
+                ChildNumber::Hardened { index: 1 },
+            ]),
+            &VERSION_MAGIC_YPUB | &VERSION_MAGIC_YPRV => Some(vec![
+                ChildNumber::Hardened { index: 49 },
+                ChildNumber::Hardened { index: 0 },
+            ]),
+            &VERSION_MAGIC_UPUB | &VERSION_MAGIC_UPRV => Some(vec![
+                ChildNumber::Hardened { index: 49 },
+                ChildNumber::Hardened { index: 1 },
+            ]),
+            &VERSION_MAGIC_ZPUB | &VERSION_MAGIC_ZPRV => Some(vec![
+                ChildNumber::Hardened { index: 84 },
+                ChildNumber::Hardened { index: 0 },
+            ]),
+            &VERSION_MAGIC_VPUB | &VERSION_MAGIC_VPRV => Some(vec![
+                ChildNumber::Hardened { index: 84 },
+                ChildNumber::Hardened { index: 1 },
+            ]),
+            &VERSION_MAGIC_ZPUB_MULTISIG
+            | &VERSION_MAGIC_ZPRV_MULTISIG
+            | &VERSION_MAGIC_YPUB_MULTISIG
+            | &VERSION_MAGIC_YPRV_MULTISIG
+                if account.is_some() =>
+            {
+                Some(vec![
+                    ChildNumber::Hardened { index: 48 },
                     ChildNumber::Hardened { index: 0 },
-                ]))
+                ])
             }
-            &VERSION_MAGIC_TPUB | &VERSION_MAGIC_TPRV => {
-                Some(DerivationPath::from(vec![
-                    ChildNumber::Hardened { index: 44 },
+            &VERSION_MAGIC_UPUB_MULTISIG
+            | &VERSION_MAGIC_UPRV_MULTISIG
+            | &VERSION_MAGIC_VPUB_MULTISIG
+            | &VERSION_MAGIC_VPRV_MULTISIG
+                if account.is_some() =>
+            {
+                Some(vec![
+                    ChildNumber::Hardened { index: 48 },
                     ChildNumber::Hardened { index: 1 },
-                ]))
-            }
-            &VERSION_MAGIC_YPUB | &VERSION_MAGIC_YPRV => {
-                Some(DerivationPath::from(vec![
-                    ChildNumber::Hardened { index: 49 },
-                    ChildNumber::Hardened { index: 0 },
-                ]))
-            }
-            &VERSION_MAGIC_UPUB | &VERSION_MAGIC_UPRV => {
-                Some(DerivationPath::from(vec![
-                    ChildNumber::Hardened { index: 49 },
-                    ChildNumber::Hardened { index: 1 },
-                ]))
-            }
-            &VERSION_MAGIC_ZPUB | &VERSION_MAGIC_ZPRV => {
-                Some(DerivationPath::from(vec![
-                    ChildNumber::Hardened { index: 84 },
-                    ChildNumber::Hardened { index: 0 },
-                ]))
-            }
-            &VERSION_MAGIC_VPUB | &VERSION_MAGIC_VPRV => {
-                Some(DerivationPath::from(vec![
-                    ChildNumber::Hardened { index: 84 },
-                    ChildNumber::Hardened { index: 1 },
-                ]))
+                ])
             }
             _ => None,
         }
+        .map(|mut path| {
+            if let Some(account_index) = account {
+                path.push(account_index);
+                match kv.as_bytes() {
+                    &VERSION_MAGIC_ZPUB_MULTISIG
+                    | &VERSION_MAGIC_ZPRV_MULTISIG
+                    | &VERSION_MAGIC_VPUB_MULTISIG
+                    | &VERSION_MAGIC_VPRV_MULTISIG => path.push(ChildNumber::Hardened { index: 2 }),
+                    &VERSION_MAGIC_YPUB_MULTISIG
+                    | &VERSION_MAGIC_YPRV_MULTISIG
+                    | &VERSION_MAGIC_UPUB_MULTISIG
+                    | &VERSION_MAGIC_UPRV_MULTISIG => path.push(ChildNumber::Hardened { index: 1 }),
+                    _ => {}
+                }
+            }
+            DerivationPath::from(path)
+        })
     }
 
     fn make_pub(kv: &KeyVersion) -> Option<KeyVersion> {
         match kv.as_bytes() {
-            &VERSION_MAGIC_XPRV => {
-                Some(KeyVersion::from_bytes(VERSION_MAGIC_XPUB))
-            }
-            &VERSION_MAGIC_YPRV => {
-                Some(KeyVersion::from_bytes(VERSION_MAGIC_YPUB))
-            }
-            &VERSION_MAGIC_ZPRV => {
-                Some(KeyVersion::from_bytes(VERSION_MAGIC_ZPUB))
-            }
-            &VERSION_MAGIC_TPRV => {
-                Some(KeyVersion::from_bytes(VERSION_MAGIC_TPUB))
-            }
-            &VERSION_MAGIC_UPRV => {
-                Some(KeyVersion::from_bytes(VERSION_MAGIC_UPUB))
-            }
-            &VERSION_MAGIC_VPRV => {
-                Some(KeyVersion::from_bytes(VERSION_MAGIC_VPUB))
-            }
+            &VERSION_MAGIC_XPRV => Some(KeyVersion::from_bytes(VERSION_MAGIC_XPUB)),
+            &VERSION_MAGIC_YPRV => Some(KeyVersion::from_bytes(VERSION_MAGIC_YPUB)),
+            &VERSION_MAGIC_ZPRV => Some(KeyVersion::from_bytes(VERSION_MAGIC_ZPUB)),
+            &VERSION_MAGIC_TPRV => Some(KeyVersion::from_bytes(VERSION_MAGIC_TPUB)),
+            &VERSION_MAGIC_UPRV => Some(KeyVersion::from_bytes(VERSION_MAGIC_UPUB)),
+            &VERSION_MAGIC_VPRV => Some(KeyVersion::from_bytes(VERSION_MAGIC_VPUB)),
             &VERSION_MAGIC_YPRV_MULTISIG => {
                 Some(KeyVersion::from_bytes(VERSION_MAGIC_YPUB_MULTISIG))
             }
@@ -629,31 +666,19 @@ impl VersionResolver for DefaultResolver {
             | &VERSION_MAGIC_YPUB_MULTISIG
             | &VERSION_MAGIC_ZPUB_MULTISIG
             | &VERSION_MAGIC_UPUB_MULTISIG
-            | &VERSION_MAGIC_VPUB_MULTISIG => Some(kv.clone()),
+            | &VERSION_MAGIC_VPUB_MULTISIG => Some(*kv),
             _ => None,
         }
     }
 
     fn make_prv(kv: &KeyVersion) -> Option<KeyVersion> {
         match kv.as_bytes() {
-            &VERSION_MAGIC_XPUB => {
-                Some(KeyVersion::from_bytes(VERSION_MAGIC_XPRV))
-            }
-            &VERSION_MAGIC_YPUB => {
-                Some(KeyVersion::from_bytes(VERSION_MAGIC_YPRV))
-            }
-            &VERSION_MAGIC_ZPUB => {
-                Some(KeyVersion::from_bytes(VERSION_MAGIC_ZPRV))
-            }
-            &VERSION_MAGIC_TPUB => {
-                Some(KeyVersion::from_bytes(VERSION_MAGIC_TPRV))
-            }
-            &VERSION_MAGIC_UPUB => {
-                Some(KeyVersion::from_bytes(VERSION_MAGIC_UPRV))
-            }
-            &VERSION_MAGIC_VPUB => {
-                Some(KeyVersion::from_bytes(VERSION_MAGIC_VPRV))
-            }
+            &VERSION_MAGIC_XPUB => Some(KeyVersion::from_bytes(VERSION_MAGIC_XPRV)),
+            &VERSION_MAGIC_YPUB => Some(KeyVersion::from_bytes(VERSION_MAGIC_YPRV)),
+            &VERSION_MAGIC_ZPUB => Some(KeyVersion::from_bytes(VERSION_MAGIC_ZPRV)),
+            &VERSION_MAGIC_TPUB => Some(KeyVersion::from_bytes(VERSION_MAGIC_TPRV)),
+            &VERSION_MAGIC_UPUB => Some(KeyVersion::from_bytes(VERSION_MAGIC_UPRV)),
+            &VERSION_MAGIC_VPUB => Some(KeyVersion::from_bytes(VERSION_MAGIC_VPRV)),
             &VERSION_MAGIC_YPUB_MULTISIG => {
                 Some(KeyVersion::from_bytes(VERSION_MAGIC_YPRV_MULTISIG))
             }
@@ -675,13 +700,15 @@ impl VersionResolver for DefaultResolver {
             | &VERSION_MAGIC_YPRV_MULTISIG
             | &VERSION_MAGIC_ZPRV_MULTISIG
             | &VERSION_MAGIC_UPRV_MULTISIG
-            | &VERSION_MAGIC_VPRV_MULTISIG => Some(kv.clone()),
+            | &VERSION_MAGIC_VPRV_MULTISIG => Some(*kv),
             _ => None,
         }
     }
 }
 
+/// Trait for building standard BIP32 extended keys from SLIP132 variant.
 pub trait FromSlip132 {
+    /// Constructts standard BIP32 extended key from SLIP132 string.
     fn from_slip132_str(s: &str) -> Result<Self, Error>
     where
         Self: Sized;
@@ -706,7 +733,7 @@ impl FromSlip132 for ExtendedPubKey {
             | VERSION_MAGIC_UPUB_MULTISIG
             | VERSION_MAGIC_VPUB_MULTISIG => VERSION_MAGIC_TPUB,
 
-            _ => Err(Error::UnknownSlip32Prefix)?,
+            _ => return Err(Error::UnknownSlip32Prefix),
         };
         data[0..4].copy_from_slice(&slice);
 
@@ -735,7 +762,7 @@ impl FromSlip132 for ExtendedPrivKey {
             | VERSION_MAGIC_UPRV_MULTISIG
             | VERSION_MAGIC_VPRV_MULTISIG => VERSION_MAGIC_TPRV,
 
-            _ => Err(Error::UnknownSlip32Prefix)?,
+            _ => return Err(Error::UnknownSlip32Prefix),
         };
         data[0..4].copy_from_slice(&slice);
 
@@ -744,3 +771,41 @@ impl FromSlip132 for ExtendedPrivKey {
         Ok(xprv)
     }
 }
+
+/// Trait converting standard BIP32 extended keys into SLIP132 representation.
+pub trait ToSlip132 {
+    /// Creates SLIP132 key representation matching the provided application
+    /// and bitcoin network.
+    fn to_slip132_string(&self, key_application: KeyApplication, network: Network) -> String;
+}
+
+impl ToSlip132 for ExtendedPubKey {
+    fn to_slip132_string(&self, key_application: KeyApplication, network: Network) -> String {
+        let key_version = DefaultResolver::resolve(network, key_application, false);
+        let mut xpub = self.encode();
+        xpub[0..4].copy_from_slice(key_version.as_slice());
+        base58::check_encode_slice(&xpub)
+    }
+}
+
+impl ToSlip132 for ExtendedPrivKey {
+    fn to_slip132_string(&self, key_application: KeyApplication, network: Network) -> String {
+        let key_version = DefaultResolver::resolve(network, key_application, true);
+        let mut xpriv = self.encode();
+        xpriv[0..4].copy_from_slice(key_version.as_slice());
+        base58::check_encode_slice(&xpriv)
+    }
+}
+
+#[cfg(test)]
+mod test {
+    use super::*;
+
+    #[test]
+    fn bip48() {
+        assert_eq!(
+            KeyApplication::from_derivation_path("m/48'/0'/8'/2'".parse().unwrap()),
+            Some(KeyApplication::SegWitMiltisig)
+        )
+    }
+}
